QQ:=Rationals();
ZZ:=Integers();
P<q>:=PuiseuxSeriesRing(QQ);

load "RankinCohen.m";
load "MoonshineCoeffs10000.m";
load "Coeffs_all.m";


function SturmBound(k,N)
	return Floor(k /12 *Index(Gamma0(N)));
end function;


function BasisToMatrix(k,N)
	M:=BaseExtend(ModularForms(N,k),QQ);	
	ind:=Index(Gamma0(N));
	bound:=SturmBound(k,N)+1;
	B:=Basis(M,bound);
	d:=#B;
	_,new_bound:=Coefficients(B[d]);
	A:=[];
	for b in B do
		coeffs,start:=Coefficients(b);
		Insert(~coeffs,1,0,[0^^start]);
		while #coeffs le new_bound do
			Append(~coeffs,0);
		end while;
		A cat:=coeffs[1..new_bound+1];
	end for;
	return Matrix(QQ,d,new_bound+1,A);
end function;


function IdentifyForm(f, k, N : test:=0)
	ind:=Index(Gamma0(N));
	bound:=SturmBound(k,N)+1;
	f+:=O(q^(bound+test));
	coeffs,start:=Coefficients(f);
	Insert(~coeffs,1,0,[0^^start]);
	prec:=AbsolutePrecision(f);
	while #coeffs lt prec do
		Append(~coeffs,0);
	end while;
	A:=BasisToMatrix(k,N);
	if #coeffs lt Ncols(A) then
		print "Not enough coefficients";
		return [];
	else
		v:=Vector(QQ,coeffs[1..Ncols(A)]);
		return Eltseq(Solution(A,v));
	end if;
end function;


function FindWeight(N)
	k:=1;
	while Dimension(CuspidalSubspace(HalfIntegralWeightForms(N,2*k-1/2))) lt 3 do
		k+:=1;
	end while;
	return 2*k-1/2;
end function;


function InPlusSpace(f)
	k:=Weight(f);
	N:=Level(f);
	bound:=SturmBound(k,Lcm(16,N));
	C,s:=Coefficients(qExpansion(f,bound));
	if IsOdd(ZZ!(k-1/2)) then
		e:=3;
	else
		e:=1;
	end if;
	return forall(t){j : j in [1..#C] | C[j] eq 0 or (j-1+s) mod 4 in [0,e] };
end function;

function Echelonize(B, k, N) // Turn a basis of a modular forms space of wt. k and Level N into echelon form
	n:=#B;
	bound:=SturmBound(k,N);
	BB:=[qExpansion(f,bound) : f in B];
	for i in [1..n] do
		min,l:=Minimum([Valuation(ff) : ff in BB[i..n]]);
		l:=l+i-1;
		ftemp:=B[i];
		fftemp:=BB[i];
		B[i]:=B[l];
		BB[i]:=BB[l];
		B[l]:=ftemp;
		BB[l]:=fftemp;
		f:=B[i];
		ff:=BB[i];
		a:=LeadingCoefficient(ff);
		m:=Valuation(ff);
		for j in [i+1..n] do
			g:=B[j];
			gg:=BB[j];
			c:=Coefficient(g,m);
			d:=GCD(a,c);
			B[j]:=(a div d)*g-(c div d)*f;
			BB[j]:=(a div d)*gg-(c div d)*ff;
		end for;
	end for;
	for i in [1..n] do
		f:=B[n-i+1];
		ff:=BB[n-i+1];
		a:=LeadingCoefficient(ff);
		m:=Valuation(ff);
		for j in [i+1..n] do
			g:=B[n-j+1];
			gg:=BB[n-j+1];
			c:=Coefficient(g,m);
			d:=GCD(a,c);
			B[n-j+1]:=(a div d)*g-(c div d)*f;
			BB[n-j+1]:=(a div d)*gg-(c div d)*ff;
		end for;
	end for;
	return B;
end function;

function QuasiEchelonize(B, k, N : bound:=SturmBound(k,N)+5 , normalize:=false) // Extends a basis of a modular forms space of wt. k and Level N to a space of QMF and returns an echelonized basis
	BB:=[Eisenstein(2,q : Precision:=bound)] cat [P!qExpansion(f,bound) : f in B];
	n:=#BB;
	for i in [1..n] do
		min,l:=Minimum([Valuation(ff) : ff in BB[i..n]]);
		l:=l+i-1;
		fftemp:=BB[i];
		BB[i]:=BB[l];
		BB[l]:=fftemp;
		ff:=BB[i];
		if ff eq O(q^bound) then
			a:=0;
		else
			a:=ZZ!LeadingCoefficient(ff);
		end if;
		m:=Valuation(ff);
		for j in [i+1..n] do
			gg:=BB[j];
			c:=ZZ!Coefficient(gg,m);
			d:=GCD(a,c);
			BB[j]:=(a div d)*gg-(c div d)*ff+O(q^bound);
		end for;
	end for;
	for i in [1..n] do
		ff:=BB[n-i+1];
		if ff eq O(q^bound) then
			a:=0;
		else
			a:=ZZ!LeadingCoefficient(ff);
		end if;
		m:=Valuation(ff);
		for j in [i+1..n] do
			gg:=BB[n-j+1];			
			c:=ZZ!Coefficient(gg,m);
			d:=GCD(a,c);
			BB[n-j+1]:=(a div d)*gg-(c div d)*ff+O(q^bound);
		end for;
	end for;
	if normalize then
		for j in [1..n] do
			f:=BB[j];	
			if not f eq O(q^bound) then	
				BB[j]:=1/LeadingCoefficient(f)*f;
			end if;
		end for;
	end if;
	return BB;
end function;



function PlusSpaceBasis(k,N) // for cusp forms only
	S:=CuspidalSubspace(HalfIntegralWeightForms(N,k));
	bound:=SturmBound(k,Lcm(16,N));
	if IsOdd(ZZ!(k-1/2)) then
		e:=3;
	else
		e:=1;
	end if;
	B:=[S.j : j in [1..Dimension(S)] | Valuation(qExpansion(S.j,bound)) mod 4 in [0,e]];
	BB:=[f : f in B | InPlusSpace(f)];
	C:=[f : f in B | not f in BB];
	while not C eq [] do
		//print C;
		f:=C[1];
		n:=Minimum([j : j in [1..bound] | not (j mod 4 in [0,e] or Coefficient(f,j) eq 0)]);
		c:=Coefficient(f,n);
		for j in [2..#C] do
			g:=C[j];
			a:=Coefficient(g,n);
			d:=GCD(a,c);
			C[j]:=S!(((c div d)*g-(a div d)*f)); 
		end for;
		Exclude(~C,f);
		//print C;
		for c in C do
			if InPlusSpace(c) then
				Append(~BB,c);
				Exclude(~C,c);
			end if;
		end for;
	end while;
	return Echelonize(BB,k,N);
end function;
		


function FindCuspForm(N)
	k:=1;
	S:=CuspidalSubspace(HalfIntegralWeightForms(N,2*k-1/2));
	bound:=SturmBound(2*k-1/2,N);
	d:=Dimension(S);
	PlusSpace:=PlusSpaceBasis(2*k-1/2,N);
	dPlus:=#PlusSpace;
	if dPlus eq 0 then
		ordmax:=0;
	else
		ordmax:=Maximum([Valuation(qExpansion(f,bound)) : f in PlusSpace]);
	end if;
	while ordmax lt 3 do
		k+:=1;
		S:=CuspidalSubspace(HalfIntegralWeightForms(N,2*k-1/2));
		bound:=SturmBound(2*k-1/2,N);
		d:=Dimension(S);
		PlusSpace:=PlusSpaceBasis(2*k-1/2,N);
		dPlus:=#PlusSpace;
		if dPlus eq 0 then
			ordmax:=0;
		else
			ordmax:=Maximum([Valuation(qExpansion(f,bound)) : f in PlusSpace]);
	end if;
	end while;
	return PlusSpace[dPlus], 2*k-1/2;
end function;


function FindCandidate(List, N : Level:=N, test:=0)
	len:=#List;
	F:=P!0;
	for n in [1..len] do
		F+:=List[n]*q^(n-4);
	end for;
	len-:=3-(len mod 4);
	F+:=O(q^len);
	//print F;
	g,k:=FindCuspForm(N);
	g:=qExpansion(g,len+8);
	//print g;
	//print F*g;
	return IdentifyForm(F*g,ZZ!(k+1/2),Level : test:=test),ZZ!(k+1/2);
end function;

function MyEisenstein(k,Prec)
	if IsOdd(k) then
		error "The weight must be even.";
	else
		return 1-2*k/BernoulliNumber(k)*&+[DivisorSigma(k-1,n)*q^n : n in [1..Prec-1]]+O(q^Prec);
	end if;
end function;

function MyDelta(Prec)
	return 1/1728*(MyEisenstein(4,Prec)^3-MyEisenstein(6,Prec)^2);
end function;


function VOperator(f,d)
	s:=Valuation(f);
	Prec:=AbsolutePrecision(f);
	return &+[Coefficient(f,n)*q^(d*n) : n in [s..Prec-1]]+O(q^(d*Prec));
end function;


function FindqExpansion(List,N,Prec : Level:=N)
	if List eq [] then // We pass on the empty list for the class 1A
		theta:=1;
		for n in [1..Floor(Sqrt(Prec+4))] do
			theta+:=2*q^(n^2);
		end for;
		theta+:=O(q^(Prec+4));
		E10:=VOperator(MyEisenstein(10,Prec),4);
		DD:=VOperator(MyDelta(Prec),4);
		return -1/10*(RankinCohenBracket(theta,1/2,E10,10,1)/DD+608*theta)+248*theta;
	else
		Coeffs,k:=FindCandidate(List,N : Level:=Level);
		M:=ModularForms(Level,k);
		f:=M!0;
		for j in [1..#Coeffs] do
			f+:=Coeffs[j]*M.j;
		end for;
		bound1:=SturmBound(k,Level);
		bound2:=SturmBound(k,N);
		F:=P!qExpansion(f,Prec+Valuation(qExpansion(f,bound1))+4);
		//print f-CuspidalProjection(f);
		g:=FindCuspForm(N);
		G:=P!qExpansion(g,Prec+Valuation(qExpansion(g,bound2))+4);
		//print G;
		return F/G;
	end if;
end function;


QQ:=Rationals();
P<q>:=PuiseuxSeriesRing(QQ);

function DifferenceOfSquares(r)	
	List:=[];
	Divs:=Divisors(r);
	for d in Divs do
		if d ge Sqrt(r) then
			Parts:=Partitions(d,2);
			for p in Parts do
				if p[1]^2-p[2]^2 eq r then
					Append(~List,p);
				end if;
			end for;
		end if;
	end for;
	return List;
end function;


function Correction(N : V:=1)	
	f:=0;
	for r in [1..N-1] do
		List:=DifferenceOfSquares(r);
		a:=0;
		for p in List do
			a+:=LegendreSymbol(p[1]*p[2],3)*(p[1]-p[2]);
		end for;
		f+:=a*q^(V*r);
	end for;
	return f+O(q^(V*N));
end function;


function Collapse(f, d)
	c,s:=Coefficients(f);
	prec:=AbsolutePrecision(f);
	supp:=[n : n in [1..#c] | not(c[n] eq 0)];
	if forall{n : n in supp | (n+s-1) mod d eq 0} then
		g:=0;	
		for n in supp do
			g+:=c[n]*q^((n+s-1) div d);
		end for;
		return g+O(q^((prec div d)+1));
	else 
		return "ERROR";
	end if;
end function;
	

function FindAllCoefficients(M);
	L:=[];
	print "Starting to compute coefficients.";
	for j in [1..48] do
		Append(~L,Coefficients(FindqExpansion(CoeffsList[j],NList[j],M+4 : Level:=LevelList[j])));
		print "Function", j, "done.";
	end for;
	max:=Maximum([#l : l in L]);
	for j in [1..#L] do
		m:=#L[j];
		if m lt max then
			L[j] cat:=[0^^(max-m)];
		end if;
	end for;
	return L;
end function;

//load "ThMatrix.m";

function MyCoeffs(f)
	cf,s:=Coefficients(f);
	prec:=RelativePrecision(f);
	return [0^^s] cat cf cat [0^^(prec-#cf)];
end function;

function sign(j)
	if j eq -3 then
		return j-1;
	else 
		return j;
	end if;
end function;



function Eta(m : Prec:=100)   // Compute the q-expansion of \eta(m\tau)
	eta:=1;
	B:=Ceiling(1/6+Sqrt(2/3*Prec+1/36))+1;
	for k in [1..B] do
		eta+:=(-1)^k*(q^(m*(k*(3*k+1)) div 2)+q^(m*(k*(3*k-1)) div 2));
	end for;
	return q^(m/24)*eta+O(q^(m*Prec));
end function;


/*function Collapse(f)
	c,s:=Coefficients(f);
	M:=AbsolutePrecision(f);
	L:=#c;
	supp:=[n+s-1 : n in [1..L] | c[n] ne 0];
	d:=GCD(supp);
	g:=O(q^(M div d));
	for n in [1..L] do
		g+:=c[n]*q^((n+s-1) div d);
	end for;
	return g,d;
end function;*/

function EtaQuotient(etalist : Prec:=100)  // Compute the q-expansion of an eta quotient, given as a list of pairs
	f:=1;
	gamma:=Matrix(2,[1,0,0,1]);
	L:=[l[2]*GCD(l[1],gamma[2][1])^2/l[1] : l in etalist | l[2] lt 0];
	if L eq [] then
		a:=0;
	else
		a:=Floor(1/24*&+L)-1;
	end if;
	
	for l in etalist do
		if l[2] ne 0 then
			f*:=(Eta(l[1] : Prec:=Ceiling((Prec-a+1)/l[1]))+O(q^(Prec-a+1)))^l[2];
		end if;
	end for;
	//print f;
	//g:=Collapse(f,24);
	if AbsolutePrecision(f) lt 0 then
		error "Precision is not high enough.";
	end if;	
	return Truncate(f)+O(q^Ceiling(AbsolutePrecision(f)));
end function;


function P1(x)
	if IsIntegral(x) then
		return 0;
	else
		return FractionalPart(x)-1/2;
	end if;
end function;


function DedekindSum(h,k)
	s:=0;
	for m in [1..k] do
		s+:=P1(h*m/k)*P1(m/k);
	end for;
	return s;
end function;


function Rademacher(g) // g=matrix([a,b;c,d]) in SL2(ZZ)
	a:=g[1][1];
	b:=g[1][2];
	c:=g[2][1];
	d:=g[2][2];
	if c eq 0 then
		return b/d;
	else 	
		return (a+d)/c-12*Sign(c)*DedekindSum(d,Abs(c));
	end if;
end function;

function SymbolUp(c,d)
	if not (GCD(c,d) eq 1 and IsOdd(d)) then
		error "Invalid arguments";
	elif c ne 0 then
		return KroneckerSymbol(c,Abs(d));
	else
		return 1;
	end if;
end function;

function SymbolDown(c,d)
	if not (GCD(c,d) eq 1 and IsOdd(d)) then
		error "Invalid arguments";
	elif c ne 0 then
		return KroneckerSymbol(c,Abs(d))*(-1)^(((Sign(c)-1)*(Sign(d)-1)) div 4);
	else
		return d;
	end if;
end function;

function EtaMultiplier(g)	
	a:=g[1][1];
	b:=g[1][2];
	c:=g[2][1];
	d:=g[2][2];
	zeta:=RootOfUnity(24);
	if IsOdd(c) then
		return SymbolUp(d,c)*zeta^((a+d)*c-b*d*(c^2-1)-3*c);
	elif IsEven(c) then
		return SymbolDown(c,d)*zeta^((a+d)*c-b*d*(c^2-1)+3*d-3-3*c*d);
	end if;
end function;
	


function EtaProductSymbol(g,etalist)     // etalist=[<d,r> : d divides Level, r exponents], <d,r> corresponds to eta(d\tau)^r.
	N:=LCM([l[1] : l in etalist]);   // the level of the eta product
	k:=&+[l[2] : l in etalist];      // the weight of the eta product is k/2
	print k;
	a:=g[1][1];
	b:=g[1][2];
	c:=g[2][1];
	d:=g[2][2];
	if c mod N ne 0 then
		printf "The matrix must be in Gamma0(%o).\n",N;
		return "Error";
	end if;
	s:=0;
	for l in etalist do
		gm:=g;
		gm[1][2]:=l[1]*b;
		gm[2][1]:=c div l[1];
		s+:=Rademacher(gm);
	end for;
	return s*k/24;
end function;


/*
load "EtaWt4/EtaWt4_56.m";
load "EtaWt4/EtaWt4_76.m";
load "EtaWt4/EtaWt4_108.m";
load "EtaWt4/EtaWt4_112.m";
load "EtaWt4/EtaWt4_128.m";
load "EtaWt4/EtaWt4_144.m";
load "EtaWt4/EtaWt4_160.m";
load "EtaWt4/EtaWt4_180.m";
load "EtaWt4/EtaWt4_216.m";
load "EtaWt4/EtaWt4_252.m";
load "EtaWt4/EtaWt4_288.m";
load "EtaWt4/EtaWt4_324.m";
load "EtaWt4/EtaWt4_360.m";
load "EtaWt4/EtaWt4_468.m";
load "EtaWt4/EtaWt4_576.m";
load "EtaWt4/EtaWt4_1152.m";
load "EtaWt4/EtaWt10_4.m";
load "EtaWt4/EtaWt6_8.m";
load "EtaWt4/EtaWt6_16.m";
load "EtaWt4/EtaWt4_12.m";
load "EtaWt4/EtaWt4_24.m";
load "EtaWt4/EtaWt4_30.m";
load "EtaWt4/EtaWt4_36.m";
load "EtaWt4/EtaWt4_40.m";
load "EtaWt4/EtaWt4_48.m";
load "EtaWt4/EtaWt4_52.m";
load "EtaWt4/EtaWt4_64.m";
load "EtaWt4/EtaWt4_72.m";
load "EtaWt4/EtaWt4_124.m";
*/

load 'EtaWt4/EtaBases.m';
load 'EtaCoeffs/EtaCoeffsAll.m';


load "etahalf.m";

function AsEtaQuotient(f, etalist, k, N) // Write a q-series f which is modular of weight k on Gamma0(N) in term of eta quotients of that weight and level
	B:=SturmBound(k,N);
	cf,start:=Coefficients(f);
	prec:=AbsolutePrecision(f);
	if prec lt B then
		error "Not enough coefficients";
	end if;
	cf:=[0^^start] cat cf;
	cf cat:=[0^^(prec-#cf)];
	v:=Vector(QQ,cf);
	A:=[];
	for eta in etalist do
		g:=EtaQuotient(eta : Prec:=prec);
		cg,sg:=Coefficients(g);
		cg:=[0^^sg] cat cg;
		if prec ge #cg then
			cg cat:=[0^^(prec-#cg)];
		else
			cg:=cg[1..prec];
		end if;
		Append(~A,cg);
	end for;
	//print v;
	A:=Matrix(QQ,A);
	print "Done building coefficient matrix, start solving linear system now";
	//print Rank(A);
	return Solution(A,v);
end function;


function ListToqSeries(List)	
	f:=P!0;
	len:=#List;
	for n in [1..len] do
		f+:=List[n]*q^(n-4);
	end for;
	return f+O(q^(len-3));
end function;


function DivideBasis(Basis,g,N) // Divide all eta quotients in Basis by eta quotient g
	NewBasis:=[];
	L:=[a[1] : a in g];
	for k in [1..#Basis] do
		b:=Basis[k];
		for i in [1..#b] do
			j:=Index(L,b[i][1]);
			if j ne 0 then
				b[i][2]-:=g[j][2];
			end if;
		end for;	
		Append(~NewBasis,b);
	end for;
	return NewBasis;
end function;


/*
for j in [1..48] do
	N:=LevelList[j];
	printf "You have chosen conjugacy class " cat NameList[j] cat ". The corresponding level is %o.\n",N;
	pdivs:=PrimeDivisors(N div 4);
	
	if 3 in pdivs then
		g:=g12;
		k1:=4;
	elif 5 in pdivs then
		g:=g20;
		k1:=4;
	elif 7 in pdivs then
		g:=g28;
		k1:=4;
	elif 13 in pdivs then
		g:=g52;
		k1:=4;
	elif 19 in pdivs then
		g:=g76;
		k1:=4;
	elif 31 in pdivs then
		g:=g124;
		k1:=4;
	elif 2 in pdivs and Valuation(N,2) le 5  then
		g:=g8;
		k1:=6;
	elif 2 in pdivs then
		g:=g32;
		k1:=4;
	else
		g:=g4;
		k1:=10;
	end if;
	
	B:=SturmBound(k1,N)+20;
	f:=ListToqSeries(CoeffsList[j][1..B+4]);
	g:=EtaQuotient(g : Prec:=B);
	
	M:=EtaBases[j];	
	
	v:=AsEtaQuotient(f*g,M,k1,N);
	
	printf "Computed eta coefficients for class %o.\n", NameList[j];	

	filename:="EtaCoeffsAll.m";
	
	Write(filename,Eltseq(v));
	Write(filename,",");
end for;
	*/
	
/*
if N eq 56 then
	g:=g28;
	printf "Please load the files 'EtaCoeffs14A.m' and 'EtaWt4/EtaWt4_56.m' to proceed.\n";
elif N eq 76 then
	g:=g76;
	printf "Please load the files 'EtaCoeffs19A.m' and 'EtaWt4/EtaWt4_76.m' to proceed.\n";
elif N eq 108 then
	g:=g12;
	printf "Please load the files 'EtaCoeffs9C.m' and 'EtaWt4/EtaWt4_108.m' to proceed.\n";
elif N eq 112 then
	g:=g28;
	printf "Please load the files 'EtaCoeffs28A.m' and 'EtaWt4/EtaWt4_112.m' to proceed.\n";
elif N eq 128 then
	g:=g32;
	printf "Please load the files 'EtaCoeffs8B.m' and 'EtaWt4/EtaWt4_128.m' to proceed.\n";
elif N eq 160 then
	g:=g32;
	printf "Please load the files 'EtaCoeffs20A.m' and 'EtaWt4/EtaWt4_160.m' to proceed.\n";
elif N eq 180 then
	g:=g12;
	printf "Please load the files 'EtaCoeffs15A.m' and 'EtaWt4/EtaWt4_180.m' to proceed.\n";
elif N eq 216 then
	g:=g12;
	printf "Please load the files 'EtaCoeffs18B.m' and 'EtaWt4/EtaWt4_216.m' to proceed.\n";
elif N eq 252 then
	g:=g28;
	printf "Please load the files 'EtaCoeffs21A.m' and 'EtaWt4/EtaWt4_252.m' to proceed.\n";
elif N eq 288 then
	g:=g12;
	printf "Please load the files 'EtaCoeffs12D.m' and 'EtaWt4/EtaWt4_288.m' to proceed.\n";
elif N eq 324 then
	g:=g12;
	printf "Please load the files 'EtaCoeffs27A.m' and 'EtaWt4/EtaWt4_324.m' to proceed.\n";
elif N eq 360 then
	g:=g12;
	printf "Please load the files 'EtaCoeffs30A.m' and 'EtaWt4/EtaWt4_360.m' to proceed.\n";
elif N eq 468 then
	g:=g52;
	printf "Please load the files 'EtaCoeffs39A.m' and 'EtaWt4/EtaWt4_468.m' to proceed.\n";
elif N eq 576 then
	g:=g12;
	printf "Please load the files 'EtaCoeffs24A.m' and 'EtaWt4/EtaWt4_576.m' to proceed.\n";
elif N eq 1152 then
	g:=g12;
	printf "Please load the files 'EtaCoeffs24C.m' and 'EtaWt4/EtaWt4_1152.m' to proceed.\n";
else
	printf "In this case everything is fine: the Maass-Poincare series cannot have a shadow.\n";
end if;
*/




printf "Please enter the number of the conjugacy class you would like to investigate.";
readi j;

	N:=LevelList[j];
	printf "You have chosen conjugacy class " cat NameList[j] cat ". The corresponding level is %o.\n",N;
	pdivs:=PrimeDivisors(N div 4);
	
	if 3 in pdivs then
		g:=g12;
		k1:=4;
	elif 5 in pdivs then
		g:=g20;
		k1:=4;
	elif 7 in pdivs then
		g:=g28;
		k1:=4;
	elif 13 in pdivs then
		g:=g52;
		k1:=4;
	elif 19 in pdivs then
		g:=g76;
		k1:=4;
	elif 31 in pdivs then
		g:=g124;
		k1:=4;
	elif 2 in pdivs and Valuation(N,2) le 5  then
		g:=g8;
		k1:=6;
	elif 2 in pdivs then
		g:=g32;
		k1:=4;
	else
		g:=g4;
		k1:=10;
	end if;
K:=CyclotomicField(24*N);
R<qq>:=PuiseuxSeriesRing(K); 

function MySqrt(n) // Compute the sqrt of an integer exactly in terms of roots of unity
	F:=Factorization(n);
	s:=1;
	for f in F do
		p:=f[1];
		e:=f[2];
		s*:=p^(e div 2);
		if p eq 2 and IsOdd(e) then
			s*:=(1+RootOfUnity(4))/RootOfUnity(8);
		elif IsOdd(p) and IsOdd(e) then
			z:=RootOfUnity(p);
			r:=0;
			for k in [1..p-1] do
				l:=k^2 mod p;
				r+:=z^l;
			end for;
			if p mod 4 eq 1 then
				r+:=1;
			else
				r:=(r+1)/RootOfUnity(4);
			end if;
			s*:=r;
		end if;
	end for;
	assert s^2 eq n;
	return s;
end function;


function EtaAtCusp(m,gamma : Prec:=20) // Compute the Fourier expansion of eta(m\tau)|gamma.
	a:=gamma[1][1];
	b:=gamma[1][2];
	c:=gamma[2][1];
	d:=gamma[2][2];
	g:=GCD(c,m);
	x:=m*a div g;
	u:=c div g;
	_,y,v:=XGCD(u,x);
	y:=-y;
	nu:=m*b*v-y*d;
	L:=Matrix(2,[x,y,u,v]);
	v:=EtaMultiplier(L);
	eta:=0;
	Prec:=Max([Prec,g^2/(24*m)+1]);
	B:=Ceiling(Sqrt(Prec*24*m)/g);
	//print B;
	for n in [1..B] do
		if GCD(12,n) eq 1 then
			eta+:=KroneckerSymbol(12,n)*qq^((n*g)^2/(24*m))*K!RootOfUnity(24*m)^ZZ!(n^2*nu*g);
		end if;
	end for;
	r:=Numerator(g/m);
	s:=Denominator(g/m);
	f:=v*MySqrt(r)/MySqrt(s)*eta+O(qq^Prec);
	if IsWeaklyZero(f) then
		error "Precision not high enough.";
	end if;
	return f;
end function;

function EtaQuotientAtCusp(etalist, gamma : Prec:=20, w:=1) // w stands for the width of the cusp
	f:=1;
	L:=[l[2]*GCD(l[1],gamma[2][1])^2/l[1] : l in etalist | l[2] lt 0];
	if L eq [] then
		a:=0;
	else
		a:=Floor(1/24*&+L)-1;
	end if;
	for l in etalist do
		if l[2] ne 0 then
			f*:=(EtaAtCusp(l[1],gamma : Prec:=(Prec+1)/w-a)+O(qq^((Prec+1)/w-a)))^l[2];
		end if;
	end for;
	//print f;
	//g:=Collapse(f,24);
	if AbsolutePrecision(f) lt 0 then
		error "Precision is not high enough.";
	end if;
	return f;
end function;

function OrdersOfEtaQuotient(etalist : Level:=N)
	HH:=UpperHalfPlaneWithCusps();
	C:=Cusps(Gamma0(N));
	orders1:=[];
	for k in C do
		_, gamma:=EquivalentPoint(HH!k);
		a:=Matrix(gamma)[2,2];
		c:=Matrix(gamma)[2,1];
		ord:=0;
		for eta in etalist do
			ord+:=GCD(c,eta[1])^2/eta[1]*eta[2];
		end for;
		fk:=EtaQuotientAtCusp(etalist,a,c : Prec:=1);
		print LeadingTerm(fk);
		Append(~orders1,<k,ord/24,QQ!Valuation(fk)>);
	end for;
	return orders1;
end function;


		


function CuspOrders(etalist,coeffs : Prec:=1)   // Compute orders at all cusps od a linear combination of eta quotients
	etalist:=DivideBasis(etalist,g,N); // Divide each basis element by the eta qutient g found above.
	HH:=UpperHalfPlaneWithCusps();
	C:=Cusps(Gamma0(NList[j]));
//	Exclude(~C,C[1]); // Exclude the cusp infinity.
	
	printf "Starting computation of expansions at cusps.\n";
	ords:=[];	
	for k in C do
		printf "Considering cusp %o.\n",k;
		_, gamma:=EquivalentPoint(HH!k);
		w:=CuspWidth(Gamma0(NList[j]),k);
		fk:=0;
		for j in [1..#coeffs] do
			if coeffs[j] ne 0 then
				printf "Starting to compute eta quotient %o out of %o.\n",j,#coeffs; 
				fk+:=coeffs[j]*EtaQuotientAtCusp(etalist[j], Matrix(gamma)^-1 : Prec:=Prec, w:=w);
			end if;
		end for;
		print fk;
		printf "The order of the given function at cusp %o is %o.\n",k,Valuation(fk);
		Append(~ords,<k,QQ!Valuation(fk)>);
	end for;
	return ords;
end function;
/*
Write("CuspOrders.m",NameList[j]);
Write("CuspOrders.m",CuspOrders(EtaBases[j],EtaCoeffsAll[j]));
Write("CuspOrders.m","\n");
*/


	
