/**************************************************************
* Functions to compute the k-th Rankin-Cohen bracket of two   *
* power series f and g with weights wf and wg.                *
**************************************************************/


//Prec:=100;

P<q>:=LaurentSeriesRing(Rationals());

			
R<spi>:=PolynomialRing(Rationals()); // spi stands for Sqrt(pi).

function MyFactorial(n)
        if IsIntegral(n)  and n ge 0 then 
		return Factorial(Integers()!n);
	elif IsIntegral(2*n) and n gt 0 then
	        f:=1;
		for l in [0..Floor(n-1/2)] do
			f*:=l+1/2;
		end for;
		return f*spi;
        elif IsIntegral(2*n) and n lt 0 then
	        m:=Floor(-n+1/2);
		f:=1;
	        for l in [0..m-2] do
			f*:=-(l+1/2);
		end for;
		return 1/f*spi;
	else
		return Gamma(n+1);
	end if;
end function;

function MyBinomial(n,k)
        if (IsIntegral(k) or IsIntegral(n-k)) and IsIntegral(2*n) then
		return Rationals()!(MyFactorial(n)/(MyFactorial(k)*MyFactorial(n-k)));
	else
		return Gamma(n+1)/(Gamma(k+1)*Gamma(n-k+1));
	end if;
end function;

function MyDerivative(f,n)
        if n eq 0 then
		return f;
	else
		return q*Derivative($$(f,n-1));
	end if;
end function;

function RankinCohenBracket(f,wf,g,wg,k)
        F:=P!0; 
	for l in [0..k] do
		F+:=(-1)^l*MyBinomial(wf+k-1,k-l)*MyBinomial(wg+k-1,l)*MyDerivative(f,l)*MyDerivative(g\
,k-l);
        end for;
	return P!F;
end function;


